from google.cloud import speech_v1p1beta1 as speech
from google.protobuf.json_format import MessageToJson

import os
import json

import ftvstt.transcripts as transcripts
import ftvstt.exceptions as exceptions
import ftvstt.transcribers as transcribers

class Google(transcribers.Transcriber):
    def __init__(self):
        super().__init__()
        self.vocList = None

    rawType = "json"

    def authenticate_with_file(self, credentialsFilePath):
        self.credentialsFilePath = credentialsFilePath

        os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = self.credentialsFilePath
        self.authed = True
        try:
            self.__client = speech.SpeechClient()
        except:
            self.authed = False

        return self.authed

    def deauthenticate(self):
        os.environ.pop('GOOGLE_APPLICATION_CREDENTIALS')
        self.credentialsFilePath = None
        self.authed = False

    @exceptions.transcribe_error_handler
    def transcribe(self, inputPath, sample_rate_hertz=16000, encoding=speech.enums.RecognitionConfig.AudioEncoding.LINEAR16, enable_word_time_offsets=True, enable_speaker_diarization=True, lang="fr-FR"):

        lang = self._make_language_compatible(lang)
        transcript = transcripts.Transcript()
        transcript.inputPath = inputPath
        transcript.provider = self.__class__
        if self.authed:
            with open(transcript.inputPath, 'rb') as inputFile:
                content = inputFile.read()

            audio = speech.types.RecognitionAudio(content=content)

            config = speech.types.RecognitionConfig(
                encoding=encoding,
                sample_rate_hertz=sample_rate_hertz,
                language_code=lang,
                enable_word_time_offsets=enable_word_time_offsets,
                enable_speaker_diarization=enable_speaker_diarization
                )

            if self.vocList is not None:
                config.speech_contexts.add(phrases=self.vocList)

            transcript.success = True
            try:
                response = self.__client.long_running_recognize(config, audio)
                resultJson = MessageToJson(response.result())
                transcript.raw = bytes(resultJson, 'ascii').decode('unicode-escape')
                self.__class__.set_text(transcript)
                self.__class__.set_words(transcript)
            except Exception as e:
                transcript.success = False
                transcript.exception = exceptions.GoogleError("An error has occured during Google Cloud transciption :\n" + str(e))

        else:
            transcript.exception = exceptions.GoogleAuthError()
            transcript.success = False
        return transcript

    def set_text(transcript):
        data = json.loads(transcript.raw)
        transcript.text = ""
        for result in data['results']:
            if 'transcript' in result['alternatives'][0].keys():
                transcript.text += result['alternatives'][0]['transcript']
        return transcript.text

    def set_words(transcript):
        transcript.words = []
        transcript.speakers = []
        data = json.loads(transcript.raw)

        for word in data['results'][-1]['alternatives'][0]['words']:
            content = word['word']
            startTime = float(word['startTime'][:-1])
            endTime = float(word['endTime'][:-1])
            speaker = transcripts.Speaker(int(word['speakerTag']))

            if not speaker in transcript.speakers:
                transcript.speakers.append(speaker)

            transcript.words.append( transcripts.Word(content, startTime=startTime, endTime=endTime, speaker=speaker) )

    def set_vocabulary_file(self, vocabularyFilePath):
        exception = None
        if self.authed:
            with open(vocabularyFilePath) as vocabularyFile:
                self.vocList = vocabularyFile.read().splitlines()
        else:
            exception = exceptions.GoogleAuthError()

        if exception:
            self.__vocList = None
            raise exception


    def _make_language_compatible(self, lang):
        if lang[2]=='-':
            lang = lang[:2].lower()+lang[2:].upper()
        return lang



import requests
import xml.etree.ElementTree as ET

import ftvstt.transcripts as transcripts
import ftvstt.exceptions as exceptions
import ftvstt.transcribers as transcribers

class Vocapia(transcribers.Transcriber):
    rawType = "xml"

    def __init__(self, apiBaseUrl):
        super().__init__()

        self.apiBaseUrl = apiBaseUrl
        self.vocabularyFilePath = None

    def __rawType(self):
        return "xml"

    def authenticate(self, user, password):
        self.user = user
        self.password = password
        self.authed = requests.post(self.apiBaseUrl, auth=(self.user, self.password)).status_code == 400
        return self.authed

    def deauthenticate(self):
        self.user = None
        self.password = None
        self.authed = False


    @exceptions.transcribe_error_handler
    def transcribe(self, inputPath, lang="fr-FR"):
        lang = self._make_language_compatible(lang)
        transcript = transcripts.Transcript()
        transcript.inputPath = inputPath
        transcript.provider = self.__class__

        if self.authed:
            payload =   {
                            'method': 'vrbs_trans',
                            'model': lang
                        }

            with open(inputPath, 'rb') as inputFile:
                if self.vocabularyFilePath is None:
                    rep = requests.post(self.apiBaseUrl , files={'audiofile': inputFile}, auth=(self.user, self.password), data=payload)
                else:
                    with open(self.vocabularyFilePath, 'rb') as vocabularyFile:
                        rep = requests.post(self.apiBaseUrl , files={'audiofile': inputFile, 'vocfile' : vocabularyFile}, auth=(self.user, self.password), data=payload)

            resultXMLroot = ET.fromstring(rep.text)
            transcript.success = resultXMLroot.tag != 'Error'
            if transcript.success:
                transcript.raw = rep.text
                self.__class__.set_text(transcript)
                self.__class__.set_words(transcript)
            else:
                transcript.exception = exceptions.VocapiaError('Vocapia error '+ resultXMLroot.attrib['code'] +' : '+  resultXMLroot.text)
        else:
            transcript.exception = exceptions.VocapiaAuthError()
            transcript.success = False
        return transcript

    def set_text(transcript):
        root = ET.fromstring(transcript.raw)
        WordsTag = root.findall('SegmentList/SpeechSegment/Word')
        Words = [word.text for word in WordsTag]
        transcript.text = ""
        for i in range(len(Words)):
            if Words[i][-2] in ["'", "-"]:
                Words[i] = Words[i][:-1]
            if i+1 < len(Words) and Words[i+1][1] in ["-",".",",","!",":","?"]:
                Words[i] = Words[i][:-1]
            transcript.text += Words[i][1:]
        return transcript.text

    def set_words(transcript):
        root = ET.fromstring(transcript.raw)
        speakers = root.findall('SpeakerList/Speaker')
        transcript.speakers = []
        for speaker in speakers:
            gender = speaker.attrib['spkid'][0]
            speakerId = int(speaker.attrib['spkid'][2:])
            transcript.speakers.append( transcripts.Speaker(speakerId,gender=gender) )

        transcript.words = []
        root = ET.fromstring(transcript.raw)
        SpeechSegments = root.findall('SegmentList/SpeechSegment')

        for speechSegment in SpeechSegments:
            gender = speechSegment.attrib['spkid'][0]
            speakerId = int(speechSegment.attrib['spkid'][2:])

            speaker = transcripts.Speaker(speakerId,gender=gender)

            words = list(speechSegment)
            for word in words:
                startTime = float(word.attrib['stime'])
                endTime = startTime + float(word.attrib['dur'])
                confidence = float(word.attrib['conf'])
                content = word.text
                transcript.words.append( transcripts.Word(content, startTime=startTime, endTime=endTime, speaker=speaker, confidence=confidence) )
        return transcript.words

    def set_vocabulary_file(self,vocabularyFilePath):
        if self.authed:
            self.vocabularyFilePath = vocabularyFilePath
        else:
            raise exceptions.VocapiaAuthError()

    def _make_language_compatible(self, lang):
        if lang.upper() in ["FR-FR","FRE","FR"]:
            return "fre"
        elif lang.upper() in ["EN-US", "EN-EN", "ENG", "EN"]:
            return "eng"
        else:
            return lang




import json

class Transcript:
    def __init__(self, raw=None, provider=None):
        if provider is None or raw is None:
            self.success = None
            self.raw = None
            self.text = None
            self.inputPath = None
            self.exception = None
            self.provider = None
            self.rawType = None
            self.exception = None

            self.words = None
            self.speakers = None
        else:
            self.raw = raw
            self.provider = provider
            self.inputPath = None
            self.rawType = provider.rawType
            provider.set_text(self)
            provider.set_words(self)
            self.success = True
            self.exception = None


    def dump_raw_result(self, outputPath):
        if self.success:
            with open(outputPath, 'w') as outputFile:
                outputFile.write(self.raw)
        else:
            raise self.exception

    def dump_text_result(self, outputPath):
        if self.success:
            with open(outputPath, 'w') as outputFile:
                outputFile.write(self.text)
        else:
            raise self.exception

    def to_dict(self):
        if self.success:
            resDict =  {
                    "provider" : self.__dict__['provider'].__name__,
                    "speakers" : [speaker.__dict__.copy() for speaker in self.__dict__['speakers']],
                    "text" : self.__dict__['text'],
                    "words" : [word.__dict__.copy() for word in self.__dict__['words']]
                    }
            for i in range(len(resDict['words'])):
                resDict['words'][i]['speakerId'] = resDict['words'][i]['speaker'].id
                del resDict['words'][i]['speaker']
            return resDict

    def dump_normalised_result(self, outputPath):
        if self.success:
            dict = self.to_dict()
            with open(outputPath, mode="w") as file:
                file.write(json.dumps(dict, ensure_ascii=False, indent=4))



class Word:
    def __init__(self, content, startTime=None, endTime=None, speaker=None, confidence=None):
        self.content = content
        self.startTime = startTime
        self.endTime = endTime
        self.speaker = speaker
        self.confidence = confidence


class Speaker:
    def __init__(self, id, gender=None):
        self.id = id
        self.gender = gender

    def __eq__(self, other) :
        if self.__class__ != other.__class__:
            return False
        return self.__dict__ == other.__dict__



import ftvstt.providers.vocapia
import ftvstt.providers.bertin
import ftvstt.providers.google
import ftvstt.providers.amazon
import ftvstt.transcripts

Vocapia = ftvstt.providers.vocapia.Vocapia
Bertin = ftvstt.providers.bertin.Bertin
Google = ftvstt.providers.google.Google
Amazon = ftvstt.providers.amazon.Amazon
Transcript = ftvstt.transcripts.Transcript



import abc
import ftvstt.exceptions as exceptions


class Transcriber(abc.ABC):
    def __init__(self):
        self.authed = False
        super().__init__()

    @abc.abstractmethod
    def authenticate(self):
        pass

    @exceptions.transcribe_error_handler
    @abc.abstractmethod
    def transcribe(self, inputPath, lang="lang"):
        pass

    @abc.abstractmethod
    def set_text(self, transcript):
        pass

    @abc.abstractmethod
    def set_words(self, transcript):
        pass

    @abc.abstractmethod
    def deauthenticate(self):
        pass



import requests
from requests.packages.urllib3.exceptions import InsecureRequestWarning
requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

import xml.etree.ElementTree as ET
import os

import ftvstt.transcripts as transcripts
import ftvstt.exceptions as exceptions
import ftvstt.transcribers as transcribers


class Bertin(transcribers.Transcriber):
    rawType = "xml"

    def __init__(self, apiBaseUrl):
        super().__init__()

        if apiBaseUrl[-1] != '/':
            self.apiBaseUrl = apiBaseUrl+'/'
        else:
            self.apiBaseUrl = apiBaseUrl

        self.__authTokenId = None
        self.__authTokenValue = None


    def authenticate(self, user, password):
        self.user = user
        self.password = password

        payload =   {
                        'login': self.user,
                        'password': self.password
                    }

        logUrl = self.apiBaseUrl + 'auth-tokens'
        rep = requests.post(logUrl, payload, verify=False)
        repDict = json.loads(rep.text)

        self.authed = rep.status_code == 201

        if self.authed:
            self.__authTokenId = repDict['id']
            self.__authTokenValue = repDict['value']

        return self.authed

    def deauthenticate(self):
        self.user = None
        self.password = None

        requests.delete(self.apiBaseUrl+"auth-tokens/" + str(self.__authTokenId),verify=False)

        self.__authTokenId = None
        self.__authTokenValue = None
        self.authed = False


    @exceptions.transcribe_error_handler
    def transcribe(self, inputPath, lang="fr-FR"):
        lang = self._make_language_compatible(lang)
        transcript = transcripts.Transcript()
        transcript.inputPath = inputPath
        transcript.provider = self.__class__
        if self.authed:
            inputName = os.path.basename(transcript.inputPath)
            lastDotIndex = inputName.rfind(".")
            apiFileName = inputName[:lastDotIndex].replace(".","_") + inputName[lastDotIndex:] #Removing dots inside the name (provoke error in Bertin)

            payload =   {
                            'audioLanguage': lang,
                            'apiFileName': apiFileName,
                            'deleteResultFilesAfterSuccess' : 'false',
                            'deleteTmpFilesAfterSuccess' : 'true',
                            'deleteIfJobFail': 'true',
                            'deleteSourceFilesAfterSuccess' : 'true',
                            'audioType':'bn',
                            'model': '',
                            'resultUrl':''
                        }

            headers =   {   'X-Auth-Token' : self.__authTokenValue,
                            "accept": "application/json"
                        }

            startTranscriptionUrl =  self.apiBaseUrl + 'transcriptions/start'
            with open(transcript.inputPath, 'rb') as inputFile:
                repJob = requests.post(startTranscriptionUrl , files={'apiFile': inputFile}, headers=headers,data=payload, verify=False)

            if repJob.status_code == 200:
                repJobDict = json.loads(repJob.text)
                jobId = repJobDict["jobId"]
                statusUrl =  self.apiBaseUrl+'transcriptions/status/'+jobId
                repStatus = requests.get(statusUrl, headers=headers, verify=False)

                if repStatus.status_code==200:
                    statusDict = json.loads(repStatus.text)

                while repStatus.status_code != 200 or statusDict['state']==1:
                    sleep(2)

                    repStatus = requests.get(statusUrl, headers=headers, verify=False)
                    if repStatus.status_code==200:
                        statusDict = json.loads(repStatus.text)
                sleep(2)
                transcript.success = statusDict['state']==3
                if transcript.success:
                    resultUrl =  self.apiBaseUrl+'transcriptions/results/'+jobId+'/xml'
                    repResult = requests.get(resultUrl, headers=headers, verify=False)
                    transcript.raw = json.loads(repResult.text)['fileContent']
                    self.__class__.set_text(transcript)
                    self.__class__.set_words(transcript)
                else:
                    transcript.exception = exceptions.BertinError("An error has occured during Bertin transciption.")
                    transcript.success = False
            else:
                transcript.exception = exceptions.BertinError("Media upload to Bertin failed.")
                transcript.success = False
        else:
            transcript.exception = exceptions.BertinAuthError()
            transcript.success = False
        return transcript

    def set_text(transcript):
        transcript.text = ""
        root = ET.fromstring(transcript.raw)
        Words = root.findall('SegmentList/SpeechSegment/Word')
        for word in Words:
            transcript.text += word.text[1:]
        return transcript.text

    def set_words(transcript):
        root = ET.fromstring(transcript.raw)
        speakers = root.findall('SpeakerList/Speaker')
        transcript.speakers = []
        for speaker in speakers:
            gender = speaker.attrib['spkid'][0]
            speakerId = int(speaker.attrib['spkid'][2:])
            transcript.speakers.append( transcripts.Speaker(speakerId,gender=gender) )

        transcript.words = []
        root = ET.fromstring(transcript.raw)
        SpeechSegments = root.findall('SegmentList/SpeechSegment')

        for speechSegment in SpeechSegments:
            gender = speechSegment.attrib['spkid'][0]
            speakerId = int(speechSegment.attrib['spkid'][2:])

            speaker = transcripts.Speaker(speakerId,gender=gender)

            words = list(speechSegment)
            for word in words:
                startTime = float(word.attrib['stime'])
                endTime = startTime + float(word.attrib['dur'])
                confidence = float(word.attrib['conf'])
                content = word.text
                transcript.words.append( transcripts.Word(content, startTime=startTime, endTime=endTime, speaker=speaker, confidence=confidence) )
        return transcript.words

    def _make_language_compatible(self, lang):
        if lang.upper() in ["FR-FR","FRE", "FR"]:
            return "fre"
        elif lang.upper() in ["EN-US", "EN-EN", "ENG", "EN"]:
            return "eng"
        elif len(lang)==3:
            return lang
        else:
            return "unknown"



import boto3
import requests

import time
import os
import json

import random as rd

import ftvstt.transcripts as transcripts
import ftvstt.exceptions as exceptions
import ftvstt.transcribers as transcribers

class Amazon(transcribers.Transcriber):
    rawType = "json"
    def __init__(self, bucketName, region_name='eu-west-3'):
        super().__init__()
        self.bucketName = bucketName
        self.region_name = region_name
        self.__vocList = None
        self.__vocName = None

    def authenticate_with_file(self, credentialsFilePath):
        self.credentialsFilePath = credentialsFilePath

        auth = {}
        self.authed = True
        try:
            with open(self.credentialsFilePath, mode="r") as csvFile:
                reader = csv.reader(csvFile, delimiter="=")
                for line in reader:
                    auth[line[0]] = line[1]

            self.__s3Client = boto3.client('s3', aws_access_key_id=auth['AWSAccessKeyId'], aws_secret_access_key=auth['AWSSecretKey'])
            self.__transcribeClient = boto3.client('transcribe', region_name=self.region_name, aws_access_key_id=auth['AWSAccessKeyId'], aws_secret_access_key=auth['AWSSecretKey'])
            client.list_buckets()

        except:
            authed = False

        return self.authed


    def authenticate(self, awsAccessKeyId, awsSecretAccessKey):
        self.id = awsAccessKeyId
        self.password = awsSecretAccessKey

        self.authed = True
        try:
            self.__s3Client = boto3.client('s3', aws_access_key_id=self.id, aws_secret_access_key=self.password)
            self.__transcribeClient = boto3.client('transcribe', region_name=self.region_name, aws_access_key_id=self.id, aws_secret_access_key=self.password)
            client.list_buckets()
        except:
            authed = False

        return self.authed

    def deauthenticate(self):
        self.credentialsFilePath = None
        self.id = None
        self.password = None

        self.__s3Client = None
        self.__transcribeClient = None


    @exceptions.transcribe_error_handler
    def transcribe(self, inputPath, lang="fr-FR", s3file=False): #set s3file=True if the input file is already on S3, in that case inputPath should be the url of the file
        lang = self._make_language_compatible(lang)
        transcript = transcripts.Transcript()
        transcript.inputPath = inputPath
        transcript.provider = self.__class__

        if self.authed:
            if s3file:
                audioUri, transcript.exception = inputPath, None
            else:
                s3AudioPath = "ftvstt/"+os.path.basename(transcript.inputPath)
                audioUri, transcript.exception = self.__s3_upload(transcript.inputPath, s3AudioPath)

            if transcript.exception:
                transcript.success = False
                return transcript

            jobName = "ftvstt_job_"+ str(rd.randint(1,1e6)) +"_"+ os.path.basename(transcript.inputPath).replace(".","_")

            try:
                if self.__vocList is None:
                    self.__transcribeClient.start_transcription_job(
                        TranscriptionJobName=jobName,
                        Media={'MediaFileUri': audioUri},
                        MediaFormat='wav',
                        LanguageCode=lang,
                        Settings={ 'MaxSpeakerLabels' : 10, 'ShowSpeakerLabels' : True}
                    )
                else:
                    self.__transcribeClient.start_transcription_job(
                        TranscriptionJobName=jobName,
                        Media={'MediaFileUri': audioUri},
                        MediaFormat='wav',
                        LanguageCode=lang,
                        Settings={ 'MaxSpeakerLabels' : 10, 'ShowSpeakerLabels' : True, 'VocabularyName': self.__vocName}
                    )
            except Exception as e:
                transcript.exception = exceptions.AmazonError("Amazon Transcription failed : \n" +str(e))
                transcript.success = False
                return transcript


            while True:
                status = self.__transcribeClient.get_transcription_job(TranscriptionJobName=jobName)
                if status['TranscriptionJob']['TranscriptionJobStatus'] in ['COMPLETED', 'FAILED']:
                    break
                time.sleep(1)

            if status['TranscriptionJob']['TranscriptionJobStatus'] == 'FAILED':
                transcript.success = False
                transcript.exception = exceptions.AmazonError("Amazon Transcription failed : \n" + status['TranscriptionJob']['FailureReason'])
            else:
                rawResultsUri = status['TranscriptionJob']['Transcript']['TranscriptFileUri']
                rep = requests.get(rawResultsUri)

                transcript.success = rep.status_code != 404

                if not transcript.success:
                    transcript.exception = exceptions.AmazonError("Amazon Transcription failed : \n" + "Impossible to reach transcription result at :" + rawResultsUri)
                    return transcript

            transcript.raw = rep.text
            self.__class__.set_text(transcript)
            self.__class__.set_words(transcript)

            time.sleep(2)
            self.__transcribeClient.delete_transcription_job(TranscriptionJobName=jobName)
        else:
            transcript.exception = exceptions.AmazonAuthError()
            transcript.success = False
        return transcript


    def set_text(transcript):
        data = json.loads(transcript.raw)
        transcript.text = ""
        for result in data['results']['transcripts']:
            transcript.text += result['transcript']
        return transcript.text

    def set_words(transcript):
        transcript.words = []
        transcript.speakers = []

        data = json.loads(transcript.raw)

        segments = data['results']['speaker_labels']['segments']

        for segment in segments:
            if not transcripts.Speaker(int(segment['speaker_label'][4:])) in transcript.speakers:
                transcript.speakers.append( transcripts.Speaker(int(segment['speaker_label'][4:])) )

        for word in data['results']['items']:
            try:
                content = word['alternatives'][0]['content']
                startTime = float(word['start_time'])
                endTime = float(word['end_time'])
                confidence = float(word['alternatives'][0]['confidence'])

                speaker = None

                for segment in segments:
                    if startTime >= float(segment['start_time']) and endTime <= float(segment['end_time']):
                        speaker = transcripts.Speaker(int(segment['speaker_label'][4:]))
                        break

                transcript.words.append( transcripts.Word(content,startTime=startTime,endTime=endTime,confidence=confidence,speaker=speaker) )
            except:
                pass


            transcript.words.append( transcripts.Word(content, startTime=startTime, endTime=endTime, confidence=confidence) )

    def __s3_upload(self, inputPath, s3Path):
        exception = None
        fileUri = None
        try:
            self.__s3Client.upload_file(inputPath, self.bucketName, s3Path)
            fileUri = "https://"+ self.bucketName +".s3."+ self.region_name +".amazonaws.com/" + s3Path
        except Exception as e:
            exception = exceptions.AmazonError("Media upload to Amazon failed : \n" +str(e))
        return fileUri, exception


    def set_vocabulary_file(self, vocabularyFilePath, lang="fr-FR"):
        exception = None
        if self.authed:
            with open(vocabularyFilePath) as vocabularyFile:
                self.__vocList = vocabularyFile.read().splitlines()
            self.__vocName = "ftvstt_voc_"+ str(rd.randint(1,1e6)) +"_"+ os.path.basename(vocabularyFilePath).replace(".","_")
            exception = self.__create_vocabulary_file(lang)
        else:
            exception = exceptions.AmazonAuthError()

        if exception:
            self.__vocList = None
            raise exception


    def __create_vocabulary_file(self, lang):
        exception = None
        self.__transcribeClient.create_vocabulary(
                VocabularyName=self.__vocName,
                LanguageCode=lang,
                Phrases=self.__vocList
            )
        while True:
            status = self.__transcribeClient.get_vocabulary(VocabularyName=self.__vocName)
            if status['VocabularyState'] in ['READY', 'FAILED']:
                break
            time.sleep(1)

        if status['VocabularyState'] == 'FAILED':
            exception = exceptions.AmazonError("Couldn't upload vocabulary file : \n" + status['FailureReason'])
        return exception

    def _make_language_compatible(self, lang):
        if lang[2]=='-':
            lang = lang[:2].lower()+lang[2:].upper()
        return lang
